import csv
import json
import pickle
import tempfile

from django.db import models, transaction
from django.core.paginator import Paginator
from django.contrib.contenttypes.models import ContentType
from django.core.files.base import File, ContentFile

from .utils import ExporterHelper


class ExporterManager(models.Manager):

    def create_exporter(self, queryset, email, attrs, limit_per_task=100):
        Model = self.model

        content = pickle.dumps(queryset.query)
        query_file = ContentFile(content)

        obj = Model.objects.create(
            content_type=ContentType.objects.get_for_model(queryset.model),
            email=email,
            attrs=json.dumps(attrs),
            limit_per_task=limit_per_task,
            total=queryset.count(),
        )

        obj.query.save(f'{obj.uuid}_queryset.pkl', query_file)

        obj.process()

        return obj

    def create_chunks_from_exporter(self, exporter, queryset):
        from .models import ExporterChunk

        chunk_list = []
        paginator = Paginator(queryset, exporter.limit_per_task)

        with transaction.atomic():
            for page in paginator.page_range:
                exporter_chunk = ExporterChunk.objects.create(exporter=exporter, page=page)
                chunk_list.append(exporter_chunk)

        return chunk_list


class ExporterChunkManager(models.Manager):

    def export(self, chunk, page_queryset, path_name, columns):
        """ Exports a CSV file containing the item_list """
        rows = []
        for obj in page_queryset:
            rows.append(ExporterHelper.get_row(obj, columns))

        with tempfile.NamedTemporaryFile(mode='w+', suffix='.csv', delete=True, encoding="utf-8") as f:
            writer = csv.writer(f, delimiter=str(';'))
            for row in rows:
                writer.writerow(row)
                f.flush()

            chunk.file.save(path_name, File(f))

        return chunk



import csv
import tempfile
import codecs

from django.core.files.storage import default_storage
from django.core.files.base import File
from django.utils.translation import ugettext as _
from model_utils.choices import Choices

from .utils import ExporterHelper


class FileHandler:
    VALID_HANDLERS = Choices(
        ("default_storage", _("default_storage"))
    )

    def __init__(self, exporter, path_name, target_storage='default_storage'):
        self.path_name = path_name
        self.exporter = exporter
        self.target = self._get_file_storage(target_storage)

    def proccess(self):
        if self.target == self.VALID_HANDLERS.default_storage:
            return self._proccess_default_storage()

    def _get_file_storage(self, storage):
        if storage not in self.VALID_HANDLERS:
            raise KeyError(_("Invalid or unsupported storage"))

        return storage

    def _proccess_default_storage(self):
        """ Join the file_list (chunked files) into one then saves and return the saved path """
        header = ExporterHelper.get_header(self.exporter.attrs)

        with tempfile.NamedTemporaryFile(mode='w+', suffix='.csv', delete=True, encoding="utf-8") as f:
            writer = csv.writer(f, delimiter=str(';'))
            writer.writerow(header)
            f.flush()

            for chunk in self.exporter.chunks.all():
                with default_storage.open(chunk.file.name) as temp_file:
                    reader = csv.reader(codecs.iterdecode(temp_file, 'utf-8'))
                    for row in reader:
                        writer.writerow(row[0].split(';'))
                        f.flush()

            self.exporter.file.save(self.path_name, File(f))

        return self.exporter



import json

from functools import reduce
from decimal import Decimal, ROUND_HALF_UP
from datetime import date, datetime


class ExporterHelper:
    """ Helper methods - This class focus on handling the files generated by the chunks/exporter """
    # TODO Refactoring is need, some code is legacy

    @classmethod
    def get_header(cls, columns):
        """ Separates the header from the json attrs"""
        columns = json.loads(columns)
        return [columns[attr] for attr in columns]

    @classmethod
    def get_row(cls, obj, attrs):
        attrs = json.loads(attrs)
        """ Generates a csv row based on the attrs and received object """
        data = cls._extract_data(obj, attrs)
        row = [data.get(attr) for attr in attrs]
        return cls._format_row(row)

    @classmethod
    def _extract_data(cls, obj, attrs):
        """ Extract data of the object based on the selected attrs """
        return {attr: cls._extract_data_by_attr(obj, attr) for attr in attrs}

    @classmethod
    def _extract_data_by_attr(cls, obj, attr, default=None):
        """ Deep searchs for the attr in the model then returns it."""
        try:
            value = cls._deepgetattr(obj, attr, default)
        except Exception as error:
            value = None
        return value

    @classmethod
    def _format_row(cls, row):
        """ Formats row for csv reading """
        _row = []
        for item in row:
            if isinstance(item, Decimal):
                item = float(item.quantize(Decimal('.01'), ROUND_HALF_UP))
            if isinstance(item, (datetime, date)):
                item = item.isoformat()
            if isinstance(item, str):
                item = str(item).replace('\n', '\r\n')
            if isinstance(item, str):
                item = item.replace('\n', '\r\n')
            _row.append(item)
        return _row

    @classmethod
    def _deepgetattr(cls, obj, attr, default=None):
        """ Deep searchs for the property on the model if exists returns it, if don't, assumes a default value """

        def getattr_(o, k):
            if hasattr(o, 'has_key'):
                v = o.get(k, default)

            #  TODO: Add tratative for int values
            else:
                v = getattr(o, k, default)

            if v:
                return v() if callable(v) else v
            else:
                return k

        return reduce(getattr_, attr.split('.'), obj)



from uuid import uuid4

from django.core.files.storage import default_storage
from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext as _

import jsonfield

from model_utils.models import TimeStampedModel
from model_utils.choices import Choices
from model_utils.fields import MonitorField

from .tasks import task_process
from .managers import ExporterManager, ExporterChunkManager
from .exceptions import ExporterStatusException


class Exporter(TimeStampedModel):
    """ Exporter model handles the initial creation of the exportation. All the info needed for the exportation comes
    from here  """

    STATUS_CHOICES = Choices(
        ("created", _("created")),
        ("processing", _("processing")),
        ("done", _("done")),
        ("error", _("error"))
    )

    uuid = models.UUIDField(default=uuid4)
    email = models.EmailField()
    content_type = models.ForeignKey(ContentType)
    query = models.FileField(blank=True, null=True, default=None, storage=default_storage)
    attrs = jsonfield.JSONField(blank=True, null=True, default=None)
    limit_per_task = models.IntegerField(default=100)
    total = models.IntegerField(editable=False)
    file = models.FileField(blank=True, null=True, default=None, storage=default_storage)
    status = models.CharField(max_length=16, choices=STATUS_CHOICES, default=STATUS_CHOICES.created)
    done_at = MonitorField(monitor='status', when=['done', 'error'], default=None, null=True)

    objects = ExporterManager()

    def set_status(self, status):
        """ Sets the status to one of the status choices """
        if status not in self.STATUS_CHOICES:
            raise ExporterStatusException("Invalid status.")

        self.status = status
        self.save()

    def process(self):
        task_process.delay(self.id)

    @property
    def chunks_is_successful(self):
        """ Check if its chunks are completed succesful """
        return self.chunks.filter(status=ExporterChunk.STATUS_CHOICES.success).count() == self.chunks.all().count()

    @property
    def has_chunk_errors(self):
        """ Check if any of the created chunks have returned errors """
        return self.chunks.filter(status=ExporterChunk.STATUS_CHOICES.error).exists()

    @property
    def is_pending(self):
        """ Check if it is pending """
        return self.status == self.STATUS_CHOICES.processing

    @property
    def is_done(self):
        return self.status == self.STATUS_CHOICES.done


class ExporterChunk(TimeStampedModel):
    STATUS_CHOICES = Choices(
        ("created", _("created")),
        ("processing", _("processing")),
        ("success", _("success")),
        ("error", _("error"))
    )

    exporter = models.ForeignKey(Exporter, related_name="chunks")
    page = models.IntegerField()
    file = models.FileField(blank=True, null=True, default=None, storage=default_storage)
    status = models.CharField(max_length=16, choices=STATUS_CHOICES, default=STATUS_CHOICES.created)
    done_at = MonitorField(monitor='status', when=['success', 'error'], default=None, null=True)

    objects = ExporterChunkManager()

    def set_status(self, status):
        """ Sets the status to one of the choices """
        if status not in self.STATUS_CHOICES:
            raise ExporterStatusException("Invalid status.")

        self.status = status
        self.save()



import os

from setuptools import setup, find_packages

from pip.req import parse_requirements

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))
__VERSION__ = '0.1.4'

base_requires = parse_requirements(os.path.join(ROOT, 'requirements', 'setup.txt'), session='hack')
base_requires = [str(item.req) for item in base_requires]

tests_requires = parse_requirements(os.path.join(ROOT, 'requirements', 'development.txt'), session='hack')
tests_requires = [str(item.req) for item in tests_requires]

setup(
    name="django-dde",
    version=__VERSION__,
    author="Stored",
    author_email="dev@stored.com.br",
    url="https://github.com/stored/django-dde",
    package_dir={'': 'src'},
    packages=find_packages('src'),
    description='Asynchronous and Distributed Data Exporter from a Django QuerySet',
    long_description=open(os.path.join(ROOT, 'README.MD'), 'r', encoding='utf8').read(),
    install_requires=base_requires,
    setup_requires=base_requires,
    extras_require={
        'tests': tests_requires,
    },
    zip_safe=True,
    include_package_data=True,
)



import pickle
import logging

from django.conf import settings
from django.utils import timezone
from django.core.paginator import Paginator
from django.core.files.storage import default_storage
from django.core.mail import send_mail

from celery import current_app as app

from .exceptions import ExporterException
from .handlers import FileHandler


logger = logging.getLogger(__name__)


@app.task(bind=True)
def task_process(self, exporter_id):
    """ After the creation of the Expoter model, this will be called to start the processing of the tasks.
    It separates the queryset in chunks for assync processing, then calls the chunk processing task """
    from .models import Exporter

    logger.info(f'[#{exporter_id}] Started exporting')

    exporter = Exporter.objects.get(pk=exporter_id)

    exporter.set_status(Exporter.STATUS_CHOICES.processing)

    Model = exporter.content_type.model_class()

    queryset = Model.objects.all()
    queryset.query = pickle.loads(exporter.query.file.read())

    chunk_list = Exporter.objects.create_chunks_from_exporter(exporter, queryset)

    logger.info(f'[#{exporter_id}] {queryset.query}')

    for chunk in chunk_list:
        task_process_chunk.delay(chunk.pk)
        logger.info(f'[#{exporter_id}] Start processing chunk page {chunk.page}')

    exporter.set_status(Exporter.STATUS_CHOICES.processing)

    logger.info(f'[#{exporter_id}] Ended processing')


@app.task(bind=True)
def task_process_chunk(self, chunk_id):
    """ Proccess the chunk from the exporter. """
    from .models import ExporterChunk

    # Fetch the Chunk model and separates the Exporter Model, also init processing for both
    exporter_chunk = ExporterChunk.objects.get(pk=chunk_id)
    exporter = exporter_chunk.exporter

    logger.info(f'[#{exporter.id}] Start processing chunk #{exporter_chunk.page}')

    exporter_chunk.set_status(ExporterChunk.STATUS_CHOICES.processing)

    Model = exporter.content_type.model_class()
    queryset = Model.objects.all()
    queryset.query = pickle.loads(exporter.query.file.read())

    # Get the current page
    paginator = Paginator(queryset, exporter.limit_per_task)
    page_queryset = paginator.page(exporter_chunk.page)

    path_name = f'reports/{exporter.uuid}/{exporter.uuid}-{exporter_chunk.page}.csv'

    try:
        ExporterChunk.objects.export(
            exporter_chunk,
            page_queryset,
            path_name,
            exporter.attrs
        )

        exporter_chunk.set_status(exporter_chunk.STATUS_CHOICES.success)

        logger.info(f'[#{exporter.id}] Ended processing chunk #{exporter_chunk.page}')

    except ExporterException as error:
        exporter_chunk.set_status(exporter_chunk.STATUS_CHOICES.error)
        logger.error(error)


@app.task(bind=True)
def task_update_exporter_status(self):
    """ Checks periodicaly if any exposter has completed. If it is then calls the finish_exporter task """
    from .models import Exporter

    # periodic_task
    logger.info('Start checking exporters status')
    exporters = Exporter.objects.filter(status=Exporter.STATUS_CHOICES.processing)

    for exporter in exporters:

        if exporter.chunks_is_successful and exporter.is_pending:
            task_finish_exporter.delay(exporter.id)
            logger.info(f'[#{exporter.id}] Finish completed')

        elif exporter.has_chunk_errors:
            exporter.set_status(Exporter.STATUS_CHOICES.error)
            logger.error(f'[#{exporter.id}] Has chunk errors')

    logger.info('Ended checking exporter status')


@app.task(bind=True)
def task_finish_exporter(self, exporter_id):
    """ Combines the chunk files in one big file and then saves """
    from .models import Exporter

    exporter = Exporter.objects.get(pk=exporter_id)

    logger.info('f[{exporter.id}] Finishing')

    now = timezone.now()

    path_name = 'reports/{date}-{uuid}.csv'.format(
        date=now.strftime('%Y-%m-%d_%H.%M.%S'),
        uuid=exporter.uuid
    )

    try:
        exporter = FileHandler(exporter, path_name).proccess()
        exporter.set_status(Exporter.STATUS_CHOICES.done)

        logger.info(f'[#{exporter.id}] Finish successful')

        task_exporter_send_email.delay(exporter_id)

    except ExporterException as error:
        exporter.set_status(exporter.STATUS_CHOICES.error)
        logger.error(error)


@app.task(bind=True)
def task_exporter_send_email(self, exporter_id):
    """ Sends a email containg the exporter file url to user """
    from .models import Exporter

    exporter = Exporter.objects.get(pk=exporter_id)
    from_email = getattr(settings, 'EXPORTER_FROM_EMAIL', [])

    if not from_email:
        logger.debug(f'[#{exporter.id}] EXPORTER_FROM_MAIL or email not set.')
        return

    send_mail(
        subject="Seu arquivo foi exportado com sucesso!",
        message=default_storage.url(str(exporter.file)),
        from_email=from_email,
        recipient_list=[exporter.email],
        fail_silently=True
    )

    logger.info(f'[#{exporter.id}] Email sent successful')
